/* ***********************************************************************/
/*                                                                       */
/*                      WiNRADiO G33DDC Excalibur Pro                    */
/*                Header file for G33DDC device driver (USB)             */
/*                     (C) Copyright 2015 RADIXON.                       */
/*                                                                       */
/* ***********************************************************************/

#ifndef __G39DDC_H__
#define __G39DDC_H__

#ifdef HAVE_CONFIG_H
#include <linux/config.h>
#endif

#define __G33_USB__
//#define __G33_PCI__
#define __G33_DEBUG__

#include <linux/version.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/time.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/kthread.h>
#include <linux/slab.h>

#ifdef CONFIG_COMPAT
#include <linux/compat.h>
#endif //CONFIG_COMPAT

#ifdef __G33_USB__
#include <linux/usb.h>
#endif //__G33_USB__

#ifdef __G33_PCI__
#include <linux/pci.h>
#include <linux/interrupt.h>
#endif //__G33_PCI__

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a<<16)|(b<<8)|c)
#endif

#if LINUX_VERSION_CODE>=KERNEL_VERSION(4,11,0)
#include <linux/sched/signal.h>
#endif

#if LINUX_VERSION_CODE<KERNEL_VERSION(2,6,27)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
typedef unsigned long uintptr_t;
#endif

#include "g33ddc_ioctl.h"
#include "g33ddc_ioctl_compat.h"

#define G33_DEVICE_NAME             "g33ddc"
#define G33_MAX_DEVICE_COUNT        32
#define G33_DESCRIPTION             "WiNRADiO G33DDC Device Driver"
#define G33_DDC1_COUNT              1

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)

#define g33_mutex                   mutex
#define g33_mutex_init(__m__)       mutex_init(__m__)
#define g33_mutex_destroy(__m__)    mutex_destroy(__m__)
#define g33_mutex_lock(__m__)       mutex_lock(__m__)
#define g33_mutex_unlock(__m__)     mutex_unlock(__m__)

#else

#define g33_mutex                   semaphore
#define g33_mutex_init(__m__)       sema_init(__m__,1)
#define g33_mutex_destroy(__m__)
#define g33_mutex_lock(__m__)       down(__m__)
#define g33_mutex_unlock(__m__)     up(__m__)

#endif

#pragma pack(push,1)

struct g33_io_buffer
{
    struct list_head        list_entry;
    uintptr_t               user_buffer;
    uint32_t                size;
    int                     page_count;
    struct page             **pages;
    int                     changed;

#ifdef __G33_USB__
    void                    *mem;
#endif //__G33_USB__

#ifdef __G33_PCI__
    struct scatterlist      *sg_list;
    int                     sg_segment_count;
    struct device           *sg_dev;
    enum dma_data_direction sg_direction;
#endif //__G33_PCI__
};

struct g33_complete_io_buffers
{
    struct list_head    list;
    struct semaphore    semaphore;
    spinlock_t          spinlock;
};

struct g33_pending_io_buffers
{
    struct list_head    list;
    spinlock_t          spinlock;
};

struct g33_io_stream
{
    uint8_t                         active;
    uint32_t                        buffer_size;
    uint32_t                        bytes_per_sec;
    struct rw_semaphore             rw_semaphore;
    struct g33_pending_io_buffers   pending_buffers;
    struct g33_complete_io_buffers  complete_buffers;
    uint16_t                        id;
};

#pragma pack(pop)

struct g33ddc;

typedef void (*g33_func_delete_t)(struct g33ddc *dev);
typedef int (*g33_func_open_t)(struct g33ddc *dev);
typedef void (*g33_func_close_t)(struct g33ddc *dev);
typedef int (*g33_func_set_power_t)(struct g33ddc *dev,uint32_t power);
typedef int (*g33_func_configure_fpga_t)(struct g33ddc *dev,void __user *user_buffer,uint32_t size);
typedef int (*g33_func_set_noise_blanker_t)(struct g33ddc *dev,uint16_t value);

typedef int (*g33_func_put_buffer_t)(struct g33ddc *dev,uint32_t stream_index,struct g33_io_buffer *io_buffer);

typedef int (*g33_func_if_start_t)(struct g33ddc *dev,uint16_t period);
typedef void (*g33_func_if_stop_t)(struct g33ddc *dev);
typedef int (*g33_func_get_adc_level_t)(struct g33ddc *dev,uint16_t *level);
typedef int (*g33_func_get_adc_sample_rate_t)(struct g33ddc *dev,uint32_t *level);

typedef int (*g33_func_ddc1_start_t)(struct g33ddc *dev,uint32_t channel,uint32_t *coeffs);
typedef void (*g33_func_ddc1_stop_t)(struct g33ddc *dev,uint32_t channel);
typedef int (*g33_func_ddc1_update_params_t)(struct g33ddc *dev,uint32_t channel);

typedef int (*g33_func_mmap_t)(struct g33ddc *dev,struct vm_area_struct *vma);
/*
typedef int (*g33_func_set_spi_regs_t)(struct g33ddc *dev,struct g33_spi_register __user *user_buffer,uint16_t count,uint16_t id);
typedef int (*g33_func_set_spi2_regs_t)(struct g33ddc *dev,struct g33_spi2_register __user *user_buffer,uint16_t count,uint16_t id);
*/
typedef int (*g33_func_rf_control)(struct g33ddc *dev,struct g33_rf_control *rf);
typedef int (*g33_func_rf_control2)(struct g33ddc *dev,struct g33_rf_control2 *rf2);

typedef int (*g33_func_usb_control1_t)(struct g33ddc *dev,const uint16_t *params,void __user *user_buffer,uint16_t buffer_size,void __user *transferred);
typedef int (*g33_func_usb_control2_t)(struct g33ddc *dev,void __user *user_buffer,uint16_t buffer_size,void __user *transferred);

typedef void (*g33_func_cancel_buffers_t)(struct g33ddc *dev,int stream_index);


struct g33ddc
{
    struct device                   dev;
    struct cdev                     cdev;
    uint32_t                        index;
    struct g33_state                state;
    struct g33_mutex                mutex;
    u64                             last_power_off_time;
    uint8_t                         dac;

    wait_queue_head_t               power_wait_queue;

    struct g33_io_stream            io_stream[G33_IO_STREAM_COUNT];
    struct g33_ddc1_params          ddc1_params[G33_DDC1_COUNT];

    g33_func_delete_t               func_delete;
    g33_func_open_t                 func_open;
    g33_func_close_t                func_close;
    g33_func_set_power_t            func_set_power;
    g33_func_configure_fpga_t       func_configure_fpga;
    /*
    g33_func_set_spi_regs_t         func_set_spi_regs;
    g33_func_set_spi2_regs_t        func_set_spi2_regs;
    */
    g33_func_rf_control             func_rf_control1;
    g33_func_rf_control2             func_rf_control2;

    g33_func_set_noise_blanker_t    func_set_noise_blanker;

    g33_func_put_buffer_t           func_put_buffer;
    g33_func_if_start_t             func_if_start;
    g33_func_if_stop_t              func_if_stop;

    g33_func_get_adc_level_t        func_get_adc_level;
    g33_func_get_adc_sample_rate_t  func_get_adc_sample_rate;

    g33_func_ddc1_start_t           func_ddc1_start;
    g33_func_ddc1_stop_t            func_ddc1_stop;
    g33_func_ddc1_update_params_t   func_ddc1_update_params;

    g33_func_cancel_buffers_t       func_cancel_buffers;

    g33_func_usb_control1_t         func_usb_control1_read;
    g33_func_usb_control1_t         func_usb_control1_write;
    g33_func_usb_control2_t         func_usb_control2_read;
    g33_func_usb_control2_t         func_usb_control2_write;
    g33_func_mmap_t                 func_mmap;

    uint8_t                         config[G33_CONFIG_SIZE];
};

#define G33_DEV_LOCK    g33_mutex_lock(&dev->mutex)
#define G33_DEV_UNLOCK  g33_mutex_unlock(&dev->mutex)

#define G33_DEV_LOCK2   g33_mutex_lock(&dev->g33_dev.mutex)
#define G33_DEV_UNLOCK2 g33_mutex_unlock(&dev->g33_dev.mutex)

#define G33_REG_LOCK    g33_mutex_lock(&g33_reg_mutex)
#define G33_REG_UNLOCK  g33_mutex_unlock(&g33_reg_mutex)

#define G33_IS_INPUT_STREAM(__index__) ((__index__)>=G33_IO_STREAM_IN_FIRST && (__index__)<=G33_IO_STREAM_IN_LAST)
#define G33_IS_OUTPUT_STREAM(__index__) ((__index__)>=G33_IO_STREAM_OUT_FIRST && (__index__)<=G33_IO_STREAM_OUT_LAST)

#define G33_ERROR(__fmt__,...) printk(KERN_ERR G33_DEVICE_NAME ": " __fmt__,##__VA_ARGS__)

#ifdef __G33_DEBUG__

#define DEBUG(__fmt__,...) printk(KERN_DEBUG G33_DEVICE_NAME ": " __fmt__,##__VA_ARGS__)

#else

#define DEBUG(__fmt__,...)

#endif //__G33_DEBUG__

#ifdef __G33_USB__

#define G33_USB_TIMEOUT                     5000

#define G33_USB_VID                         0x14E0
#define G33_USB_PID                         0x1103

#define G33_USB_EP_IN_FPGA                  0x02
#define G33_USB_EP_IN_CTRL                  0x01
#define G33_USB_EP_OUT_CTRL                 0x06
#define G33_USB_EP_OUT_FPGA                 0x08


#define USB_REG_CTRL1                   0x0000
#define USB_REG_FLUSH                   0x0001
#define USB_REG_COMMAND_ID              0x0002
#define USB_REG_SPI_LO                  0x0004
#define USB_REG_SPI_HI                  0x0005
#define USB_REG_IF_BUFF_CTRL            0x0008
#define USB_REG_NOISE_BLANKER_THRESHOLD 0x0009
#define USB_REG_COEFF_WR_CTRL           0x000C
#define USB_REG_DDC1_FIR_COEFF_0        0x000D
#define USB_REG_DDC1_FIR_COEFF_1        0x000E
#define USB_REG_DDC2_FIR_1_COEFF_0      0x000F
#define USB_REG_DDC2_FIR_1_COEFF_1      0x0010
#define USB_REG_DDC2_FIR_2_COEFF_0      0x0011
#define USB_REG_DDC2_FIR_2_COEFF_1      0x0012
#define USB_REG_DDC1_CTRL_0             0x0018
#define USB_REG_DDC1_TUNING_WORD_0      0x0019
#define USB_REG_DDC1_CIC_CTRL_0         0x001B
#define USB_REG_DDC1_FIR_CTRL_0         0x001C
#define USB_REG_DDC1_CTRL_1             0x0020
#define USB_REG_DDC1_TUNING_WORD_1      0x0021
#define USB_REG_DDC1_CIC_CTRL_1         0x0023
#define USB_REG_DDC1_FIR_CTRL_1         0x0024

#define USB_REG_DDC2_SHIFT_CONTROL_0    0x0028
#define USB_REG_DDC2_SHIFT_TUNING_0     0x0029
#define USB_REG_DDC2_FIR1_CONTROL1_0    0x002A
#define USB_REG_DDC2_FIR1_CONTROL2_0    0x002B
#define USB_REG_DDC2_FIR2_CONTROL1_0    0x002C
#define USB_REG_DDC2_FIR2_CONTROL2_0    0x002D
#define USB_REG_DDC2_RETIMER_0          0x002E
#define USB_REG_DDC2_INPUT_SWITCH_0     0x002F

#define USB_REG_DDC2_SHIFT_CONTROL_1    0x0032
#define USB_REG_DDC2_SHIFT_TUNING_1     0x0033
#define USB_REG_DDC2_FIR1_CONTROL1_1    0x0034
#define USB_REG_DDC2_FIR1_CONTROL2_1    0x0035
#define USB_REG_DDC2_FIR2_CONTROL1_1    0x0036
#define USB_REG_DDC2_FIR2_CONTROL2_1    0x0037
#define USB_REG_DDC2_RETIMER_1          0x0038
#define USB_REG_DDC2_INPUT_SWITCH_1     0x0039

#define USB_REG_DAC_CONTROL             0x003C

#define G33_USB_INPUT_BUFFER_MIN_SIZE       (8*1024)
#define G33_USB_INPUT_BUFFER_MAX_SIZE       (126*1024)
#define G33_USB_INPUT_BUFFER_COUNT          4
#define G33_USB_INPUT_SWP_BUFFER_SIZE       (16384)
#define G33_USB_OUTPUT_BUFFER_COUNT         4
#define G33_USB_OUTPUT_BUFFER_SIZE          (64*1024)

#define G33_USB_INPUT_PACKET_PAYLOAD_SIZE   (96)

#pragma pack(push,1)

struct g33_usb_input_packet
{
    uint8_t     sync[2];
    uint8_t     source;
    uint8_t     flags;
    uint16_t    id;
    uint8_t     data[G33_USB_INPUT_PACKET_PAYLOAD_SIZE];
};

struct g33_usb_output_packet
{
    uint16_t    sync;
    uint16_t    length;
    uint16_t    command;
    uint16_t    address;
    union
    {
        uint8_t     data8[0];
        uint16_t    data16[0];
        uint32_t    data32[0];
    };
};

#pragma pack(pop)

struct g33ddc_usb;

struct g33_usb_buffer
{
    struct urb              *urb;
    uint8_t                 *mem;
    struct list_head        list_entry;
    uint8_t                 pending;
    struct g33ddc_usb       *dev;
};

struct g33_usb_buffers
{
    struct list_head    list;
    spinlock_t          spinlock;
};

struct g33ddc_usb
{
    struct g33ddc                   g33_dev;
    struct usb_interface            *usb_interface;
    struct usb_device               *usb_dev;
    uint16_t 						IFCommandId;
    uint16_t						ADC_Level;
    uint32_t						TCXO_Freq;

    struct
    {
        struct task_struct          *thread;
        uint32_t                    stream_counter;
        struct g33_usb_buffer       buffer[G33_USB_INPUT_BUFFER_COUNT];
        struct g33_usb_buffers      complete_usb_buffers;
        uint32_t                    buffer_size;
    }                               input;

    struct
    {
        struct task_struct          *thread;
        uint32_t                    stream_counter;
        struct g33_usb_buffer       buffer[G33_USB_OUTPUT_BUFFER_COUNT];
        struct g33_usb_buffers      avail_usb_buffers;
        uint32_t                    bytes_per_packet[G33_IO_STREAM_OUT_COUNT];
        uint32_t                    ddc1_divider;
        uint8_t                     dac_control;
    }                               output;

    struct g33_io_buffer            *io_buffer[G33_IO_STREAM_COUNT];

};

#endif //__G33_USB__


#endif //__G39DDC_H__
